package org.simplericity.jettyconsole.jsp;

import org.apache.jasper.servlet.JspServlet;
import org.eclipse.jetty.webapp.AbstractConfiguration;
import org.eclipse.jetty.webapp.WebAppContext;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Glassfish JSP / Jasper is too jar files based in a few cases.
 * This Jetty configuration postConfigures the Jasper JSP servlet
 * to not use redeployMode for Jasper's TldLocationsCache.
 *
 * We also preemptively run TldLocationsCache's init method, collect the
 * <code>mappings</code> map and replace all <code>location</code> values
 * such that TLD's under /META-INF/jettyconsole/lib/* can be found by Jasper's.
 * TagLibraryInfoImpl.
 *
 * All of this done using reflection. Ugly bugly.
 *
 */
public class JasperInUnpackedWarUglyHackConfiguration extends AbstractConfiguration {

    public static final String JSP_TLD_URI_TO_LOCATION_MAP =
        "com.sun.jsp.tldUriToLocationMap";

    @Override
    public void postConfigure(WebAppContext webAppContext) throws Exception {

        System.setProperty("org.apache.jasper.compiler.disablejsr199", "true");

        // Grab the JSP servlet
        JspServlet servlet = (JspServlet) webAppContext.getServletHandler().getServlet("jsp").getServletInstance();

        // Get the EmbeddedServletOptions options object
        Field optionsField = JspServlet.class.getDeclaredField("options");
        optionsField.setAccessible(true);
        Object optionsObject = optionsField.get(servlet);

        // Get the TldLocationsCache tldLocationsCache object
        Field cacheField = optionsObject.getClass().getDeclaredField("tldLocationsCache");
        cacheField.setAccessible(true);
        Object cacheObject = cacheField.get(optionsObject);


        // Set TldLocationsCache.redeployMode to false
        Field redeployModeField = cacheObject.getClass().getDeclaredField("redeployMode");
        redeployModeField.setAccessible(true);
        redeployModeField.set(cacheObject, Boolean.FALSE);

        // Run TldLocationsCache.init()
        ClassLoader contextLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(webAppContext.getClassLoader());
        try {
            Method initMethod = cacheObject.getClass().getDeclaredMethod("init");
            initMethod.setAccessible(true);
            initMethod.invoke(cacheObject);

        } finally {
            if(contextLoader != null) {
                Thread.currentThread().setContextClassLoader(contextLoader);
            }
        }

        Field mappingsField = cacheObject.getClass().getDeclaredField("mappings");
        mappingsField.setAccessible(true);

        Map<String, String[]> mappings = (Map<String, String[]>) mappingsField.get(cacheObject);


        Method getUriFromTld = cacheObject.getClass().getDeclaredMethod("getUriFromTld", String.class, InputStream.class);
        getUriFromTld.setAccessible(true);

        final String warFile = getWarPath();
        for(String path : getJettyConsoleTldPaths()) {
            String resourcePath = "/META-INF/jettyconsole/lib" + path;
            URL resourceUrl = new URL("jar:" + warFile +"!" + resourcePath);
            String uri = (String) getUriFromTld.invoke(cacheObject, resourceUrl.toString(), resourceUrl.openStream());
            if(uri != null && !mappings.containsKey(uri)) {
                mappings.put(uri, new String[] {resourcePath, null});
            }
        }
    }

    /**
     * This configuration might be reused for other webappcontexts than JettyConsole's own, so
     * make sure we read from the right path.
     * @return set of TLDs bundled with JettyConsole
     */
    private Set<String> getJettyConsoleTldPaths() {
        final String warFile = getWarPath();

        try {
            URL pathUrl = new URL("jar:" + warFile + "!" + "/META-INF/jettyconsole/jettyconsolepaths.txt");

            BufferedReader reader = new BufferedReader(new InputStreamReader(pathUrl.openStream()));
            String line;
            Set<String> tldPaths = new HashSet<String>();
            while(( line = reader.readLine()) != null) {
                if(line.contains("/META-INF/") && line.endsWith(".tld")) {
                    tldPaths.add(line);
                }
            }
            reader.close();
            return tldPaths;
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private String getWarPath() {
        final URL classResource = getClass().getResource(getClass().getSimpleName() + ".class");
        final String file = URLDecoder.decode(classResource.getFile());

        return file.substring(0, file.indexOf("!"));
    }
}
